
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layLayoutView.h"
#include "laySaveLayoutOptionsDialog.h"
#include "layStream.h"
#include "layExceptions.h"
#include "dbSaveLayoutOptions.h"
#include "dbStream.h"
#include "tlClassRegistry.h"

#include <QtGui/QScrollArea>

namespace lay
{

static const StreamWriterPluginDeclaration *plugin_for_format (const std::string &format_name)
{
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    const StreamWriterPluginDeclaration *decl = dynamic_cast <const StreamWriterPluginDeclaration *> (&*cls);
    if (decl && decl->format_name () == format_name) {
      return decl;
    }
  }
  return 0;
}

SaveLayoutOptionsDialog::SaveLayoutOptionsDialog (QWidget *parent, const std::string &title)
  : QDialog (parent, "save_layout_options_dialog"), Ui::SaveLayoutOptionsDialog (),
    mp_config_root (0)
{
  Ui::SaveLayoutOptionsDialog::setupUi (this);

  setCaption (title.c_str ());

  QWidget *empty_widget = new QWidget (options_stack);
  int empty_widget_index = options_stack->addWidget (empty_widget);

  for (tl::Registrar<db::StreamFormatDeclaration>::iterator fmt = tl::Registrar<db::StreamFormatDeclaration>::begin (); fmt != tl::Registrar<db::StreamFormatDeclaration>::end (); ++fmt) {

    fmt_cbx->addItem (fmt->format_desc ().c_str ());

    ConfigPage *page = 0;

    //  obtain the config page from the plugin which we identify by format name
    const StreamWriterPluginDeclaration *decl = plugin_for_format (fmt->format_name ());
    if (decl) {
      std::string title; // ignored
      page = decl->config_page (options_stack, title);
    }

    m_pages.push_back (std::make_pair (page, page ? options_stack->addWidget (page) : empty_widget_index));

  }

  connect (buttonBox, SIGNAL (accepted ()), this, SLOT (ok_button_pressed ()));
  connect (fmt_cbx, SIGNAL (activated (int)), this, SLOT (fmt_cbx_changed (int)));
}

SaveLayoutOptionsDialog::~SaveLayoutOptionsDialog ()
{
  //  .. nothing yet ..
}

void
SaveLayoutOptionsDialog::ok_button_pressed ()
{
  BEGIN_PROTECTED

  for (std::vector< std::pair<ConfigPage *, int> >::iterator page = m_pages.begin (); page != m_pages.end (); ++page) {
    if (page->first) {
      page->first->commit (mp_config_root);
    }
  }

  double x = 0.0;
  tl::from_string (dbu_le->text ().ascii (), x);
  tl::from_string (sf_le->text ().ascii (), x);

  accept ();

  END_PROTECTED
}

bool 
SaveLayoutOptionsDialog::get_options (lay::LayoutView *view, lay::PluginRoot *config_root, unsigned int cv_index, const std::string &fn, bool &gzip, db::SaveLayoutOptions &options)
{
  filename_lbl->setText (fn.c_str ());
  gzip_cb->setChecked (gzip);

  dbu_le->setText (tl::to_string (options.dbu ()).c_str ());

  unsigned int i = 0;
  for (tl::Registrar<db::StreamFormatDeclaration>::iterator fmt = tl::Registrar<db::StreamFormatDeclaration>::begin (); fmt != tl::Registrar<db::StreamFormatDeclaration>::end (); ++fmt, ++i) {
    if (fmt->format_name () == options.format ()) {
      fmt_cbx->setCurrentIndex (i);
      fmt_cbx_changed (i);
      break;
    }
  }

  mp_config_root = config_root;

  for (std::vector< std::pair<ConfigPage *, int> >::iterator page = m_pages.begin (); page != m_pages.end (); ++page) {
    if (page->first) {
      page->first->setup (config_root);
    }
  }

  bool ret = false;

  if (exec ()) {

    gzip = gzip_cb->isChecked ();

    int index = fmt_cbx->currentIndex ();
    for (tl::Registrar<db::StreamFormatDeclaration>::iterator fmt = tl::Registrar<db::StreamFormatDeclaration>::begin (); fmt != tl::Registrar<db::StreamFormatDeclaration>::end () && index >= 0; ++fmt, --index) {
      if (index == 0) {
        options.set_format (fmt->format_name ());
      }
    }

    double dbu = 0.0;
    tl::from_string (dbu_le->text ().ascii (), dbu);

    double sf = 1.0;
    tl::from_string (sf_le->text ().ascii (), sf);

    options.set_dbu (dbu);
    options.set_scale_factor (sf);

    if (layersel_cbx->currentIndex () == 0 /*all*/) {
      options.select_all_layers ();
    } else if (layersel_cbx->currentIndex () == 1 /*shown layers*/) {
      options.deselect_all_layers ();
      for (LayerPropertiesConstIterator layer = view->begin_layers (); layer != view->end_layers (); ++layer) {
        if (layer->cellview_index () == int (cv_index)) {
          options.add_layer (layer->layer_index ());
        }
      }
    } else if (layersel_cbx->currentIndex () == 2 /*visible layers*/) {
      options.deselect_all_layers ();
      for (LayerPropertiesConstIterator layer = view->begin_layers (); layer != view->end_layers (); ++layer) {
        if (layer->cellview_index () == int (cv_index) && layer->visible (true)) {
          options.add_layer (layer->layer_index ());
        }
      }
    }

    //  identify the plugin in charge of creating the particular options
    const StreamWriterPluginDeclaration *decl = plugin_for_format (options.format ());
    if (decl) {
      options.set_specific_options (decl->create_specific_options (config_root));
    }

    ret = true;

  }

  mp_config_root = 0;
  return ret;
}

void 
SaveLayoutOptionsDialog::fmt_cbx_changed (int index)
{
  if (index >= 0 && index < int (m_pages.size ())) {
    options_stack->setCurrentIndex (m_pages[index].second);
  }
}

}

